from statistics import mean, stdev
import logging
from ....models.hourselection.cautionhourtype import CautionHourType, MAX_HOURS

_LOGGER = logging.getLogger(__name__)


def normalize_prices(prices: list) -> list:
    min_price = min(prices)
    c = 0
    if min_price <= 0:
        c = abs(min_price) + 0.01
    ret = []
    for p in prices:
        pp = p + c
        divider = min_price if min_price > 0 else c
        ret.append(round(pp - divider, 3))
    return ret


def get_offset_dict(normalized_hours: list):
    ret = {}
    _prices = [p - min(normalized_hours) for p in normalized_hours]
    average_val = mean(_prices)
    for i in range(0, len(_prices)):
        # todo: fix this to accommodate quarters also
        try:
            ret[i] = round((_prices[i] / average_val) - 1, 2)
        except:
            ret[i] = 1
    return ret


async def async_create_cautions(
    hourdict: dict,
    normalized_hourdict: dict,
    cautionhour_type: CautionHourType,
    range_start: int = 0,
    adjusted_average: float | None = None,
    blocknocturnal: bool = False,
) -> dict:
    """Rank the normalized pricelist to find out which are going to become non- or caution-hours"""

    _adj_avg = adjusted_average
    if not isinstance(adjusted_average, (float, int)):
        _adj_avg = mean(hourdict.values())

    adj_average_norm = _adj_avg * (
        mean(normalized_hourdict.values()) / mean(hourdict.values())
    )
    cautions = [
        h
        for h in normalized_hourdict
        if normalized_hourdict[h] > (adj_average_norm * 0.7)
    ]
    cautions_dict = await async_cap_pricelist_available_hours(
        cautions, normalized_hourdict, cautionhour_type, blocknocturnal, range_start
    )
    maxval = max(hourdict.values())
    ret = {}

    try:
        for k, v in cautions_dict.items():
            ret[k] = {
                "val": hourdict[k],
                "permax": round(hourdict[k] / maxval, 2),
                "force_non": v,
            }
    except IndexError as e:
        _LOGGER.error(f"Error on creating the cautions_dict: {e}")

    if blocknocturnal:
        return ret
    else:
        return await async_discard_excessive_hours(ret)


async def async_cap_pricelist_available_hours(
    cautions: list,
    normalized_hourdict: dict,
    cautionhour_type: CautionHourType,
    blocknocturnal: bool,
    range_start: int,
) -> dict:
    _demand = max(len(normalized_hourdict.keys()) - MAX_HOURS.get(cautionhour_type), 0)
    ret = {c: False for c in cautions} if _demand == 0 else {c: True for c in cautions}
    hours_sorted = [
        k for k, v in sorted(normalized_hourdict.items(), key=lambda item: item[1])
    ]
    iterations = 0
    while len(cautions) < _demand and iterations < len(hours_sorted) * 2:
        iterations += 1
        idx = max(range_start - 1, 0)
        if len(cautions) > 0:
            idx = hours_sorted.index(cautions[-1])

        if idx <= len(hours_sorted) - 1:
            try:
                while idx + 1 < len(hours_sorted):
                    idx += 1
                    next = hours_sorted[idx]
                    if next not in cautions:
                        cautions.append(next)
                        ret[next] = True
                        break
            except IndexError as e:
                raise IndexError(f"error on first. idx:{idx}")
        if idx >= 0:
            try:
                while idx - 1 >= 0:
                    idx -= 1
                    prev = hours_sorted[idx]
                    if prev not in cautions:
                        cautions.append(prev)
                        ret[prev] = True
                        break
            except IndexError as e:
                raise IndexError(f"error on second. idx:{idx}")

    try:
        for i in await async_get_nocturnal_stop(blocknocturnal, range_start):
            if i not in cautions:
                ret[i] = True
    except IndexError as e:
        _LOGGER.error(f"error on looping nocturnal stop: {e}")
    return await async_sort_by_key(ret)


async def async_sort_by_key(input: dict) -> dict:
    _keys = list(input.keys())
    _keys.sort()
    return {i: input[i] for i in _keys}


async def async_get_nocturnal_stop(
    blocknocturnal: bool = False, range_start: int = 0
) -> list:
    _base = [23, 0, 1, 2, 3, 4, 5, 6]
    if blocknocturnal:
        return await async_transform_range(range_start, _base)
    return []


async def async_transform_range(range_start, base):
    if range_start == 0:
        return base
    ret = []
    for b in base:
        if b >= range_start:
            ret.append(b - range_start)
        else:
            ret.append(24 - range_start + b)
    return ret


async def async_discard_excessive_hours(hours: dict):
    """There should always be at least four regular hours before absolute_top_price kicks in."""
    while len(hours) >= 20:
        to_pop = dict(sorted(hours.items(), key=lambda item: item[1]["val"]))
        hours.pop(list(to_pop.keys())[0])
    return hours


async def async_should_be_cautionhour(
    price_item, prices, peak, cautionhour_type
) -> bool:
    first = any(
        [
            float(price_item["permax"]) <= cautionhour_type,
            float(price_item["val"]) <= (sum(prices) / len(prices)),
        ]
    )
    second = (peak > 0 and peak * price_item["permax"] > 1) or peak == 0
    return all([first, second])


async def async_set_charge_allowance(price_input, cautionhour_type) -> float:
    return round(abs(price_input - 1), 2) * ALLOWANCE_SCHEMA[cautionhour_type]


ALLOWANCE_SCHEMA = {
    CautionHourType.get_num_value(CautionHourType.SUAVE): 1.15,
    CautionHourType.get_num_value(CautionHourType.INTERMEDIATE): 1.05,
    CautionHourType.get_num_value(CautionHourType.AGGRESSIVE): 1,
    CautionHourType.get_num_value(CautionHourType.SCROOGE): 1,
}



from __future__ import annotations
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from ....models.hourselection.hourselection_options import HourSelectionOptions
from dataclasses import dataclass, field
from .hour_price import HourPrice
from datetime import date, datetime, time
from .list_type import ListType


@dataclass
class HourSelectionModel:
    prices_today: list[float] = field(default_factory=list)
    prices_tomorrow: list[float] = field(default_factory=list)
    _hours_prices: list[HourPrice] = field(default_factory=list)
    adjusted_average: float | None = None

    @property
    def hours_prices(self) -> list[HourPrice]:
        return self._hours_prices

    @hours_prices.setter
    def hours_prices(self, val: list[HourPrice]) -> None:
        if val is not None:
            if len(val) > 0:
                self._hours_prices = val
                self._hours_prices.sort(key=lambda x: x.dt)
        else:
            self._hours_prices = []

    def set_hourprice_list(
        self,
        prices: list,
        service_options: HourSelectionOptions,
        is_quarterly: bool,
        datum: date,
        is_passed_func: Callable,
    ) -> None:
        ret = []
        for idx, p in enumerate(prices):  # type: ignore
            assert isinstance(p, (float, int))
            hour = int(idx / 4) if is_quarterly else idx
            quarter = idx % 4 if is_quarterly else 0
            ret.append(
                HourPrice(
                    dt=self._get_dt(datum, hour, quarter),
                    quarter=quarter,
                    price=p,
                    passed=is_passed_func(datum, hour, quarter),
                    hour_type=HourPrice.set_hour_type(
                        service_options.absolute_top_price, service_options.min_price, p
                    ),
                    list_type=ListType.Quarterly if is_quarterly else ListType.Hourly,
                )
            )
        if len(self.hours_prices) > 0:
            self.hours_prices.extend(ret)
        else:
            self.hours_prices = ret

    @staticmethod
    def _get_dt(datum: date, hour: int, quarter: int) -> datetime:
        return datetime.combine(
            datum, time(hour=hour, minute=quarter * 15, second=0, microsecond=0)
        )



from dataclasses import dataclass, field
from .hour_price import HourPrice
from datetime import datetime


@dataclass
class MaxMinModel:
    min_price: float
    input_hours: list[HourPrice] = field(default_factory=lambda: [])
    original_input_hours: list[HourPrice] = field(default_factory=lambda: [])
    expected_hourly_charge: float = 0

    def caluclate_average_price(
        self, input: list[HourPrice], total_charge: float | None, dt: datetime
    ) -> float | None:
        _total = total_charge or 0
        _dt = dt.replace(minute=0, second=0, microsecond=0)
        try:
            first = sum(
                [
                    hp.permittance * hp.price * self.expected_hourly_charge
                    for hp in input
                    if hp.permittance > 0 and hp.dt >= _dt
                ]
            )
            return round(first / _total, 2)
        except ZeroDivisionError as e:
            return 0.0

    def calculate_total_charge(self, input: list[HourPrice], dt: datetime) -> float:
        _dt = dt.replace(minute=0, second=0, microsecond=0)
        return round(
            sum(
                [
                    hp.permittance * self.expected_hourly_charge
                    for hp in input
                    if hp.dt >= _dt
                ]
            ),
            1,
        )



from .models.hour_price import HourPrice
from ...models.hourselection.cautionhourtype import CautionHourType
from .models.hour_type import HourType
from statistics import mean


def set_initial_permittance(
    hours: list[HourPrice],
    price_mean: float,
    price_stdev: float,
    avg7: float | None = None,
) -> None:
    scaling_factor = scale_permittance(price_stdev)
    for hour in hours:
        if hour.hour_type == HourType.BelowMin:
            hour.permittance = 1.0
        elif hour.hour_type == HourType.AboveMax:
            hour.permittance = 0.0
        else:
            diff_avg_price = price_mean - hour.price
            diff_avg_7day_price = (avg7 or price_mean) - hour.price
            hour.permittance = mean(
                [
                    max(0, min(1, diff_avg_price / (price_mean * scaling_factor))),
                    max(
                        0,
                        min(
                            1,
                            diff_avg_7day_price
                            / ((avg7 or price_mean) * scaling_factor),
                        ),
                    ),
                ]
            )


def scale_permittance(stdev: float) -> float:
    if stdev == 0:
        return 1.0
    return max(0.1, min(1.0, 1.0 / (1.0 + stdev)))


def set_scooped_permittance(
    hour_prices: list[HourPrice], caution_hour_type: CautionHourType
) -> None:
    lo_cutoff = 0.4
    hi_cutoff = 0.75
    max_hours = 24  # todo: add support for 96 if quarterly
    min_hours = 4
    match caution_hour_type:
        case CautionHourType.SUAVE:
            hi_cutoff = 0.7
        case CautionHourType.INTERMEDIATE:
            lo_cutoff = 0.5
            hi_cutoff = 0.7
        case CautionHourType.AGGRESSIVE:
            lo_cutoff = 0.6
        case CautionHourType.SCROOGE:
            lo_cutoff = 0.6
            max_hours = 8  # todo: add support for 32 if quarterly
            min_hours = 0

    for hp in hour_prices:
        if hp.permittance <= lo_cutoff:
            hp.permittance = 0.0
        elif hp.permittance >= hi_cutoff:
            hp.permittance = 1.0
        else:
            hp.permittance = round(hp.permittance, 2)

    if len([hp for hp in hour_prices if hp.permittance > 0.0]) < min_hours:
        _t = [hp for hp in hour_prices if hp.list_type != HourType.AboveMax]
        if len(_t):
            _t.sort(key=lambda x: x.price)
            for h in range(min_hours):
                _t[h].permittance = 1.0 if _t[h].permittance == 0 else _t[h].permittance



from .models.stop_string import AllowanceObj, set_allowance_obj
from .models.datetime_model import DateTimeModel
from .models.hour_price import HourPrice
from .models.list_type import ListType
from .models.hourselection_model import HourSelectionModel
from statistics import stdev, mean
from datetime import date, datetime, time
from ...models.hourselection.hourselection_options import HourSelectionOptions
from .hourselection_calculations import normalize_prices
from .offset_dict import get_offset_dict, set_offset_dict
from .permittance import set_initial_permittance, set_scooped_permittance
from .max_min_charge import MaxMinCharge


class HourSelectionService:
    def __init__(self, options: HourSelectionOptions = HourSelectionOptions()):
        self.options = options
        self.use_quarters: bool = False
        self.dtmodel = DateTimeModel()
        self.model = HourSelectionModel()
        self.max_min = MaxMinCharge(service=self, min_price=self.options.min_price)
        self._offset_dict: dict[datetime, dict] = {}

    @property
    def all_hours(self) -> list[HourPrice]:
        self.update()
        return self.model.hours_prices

    @property
    def future_hours(self) -> list[HourPrice]:
        self.update()
        return self.get_future_hours()

    @property
    def passed_hours(self) -> list[HourPrice]:
        self.update()
        return [hp for hp in self.model.hours_prices if hp.passed]

    @property
    def stopped_string(self) -> str:
        self.update()
        return self.allowance.display_name

    @property
    def allowance(self) -> AllowanceObj:
        return set_allowance_obj(self.dtmodel, self.future_hours)

    @property
    def average_kwh_price(self) -> float:
        return self.get_average_kwh_price()

    @property
    def offset_dict(self) -> dict:
        return get_offset_dict(self._offset_dict)

    def update(self):
        for hp in self.model.hours_prices:
            hp.set_passed(self.dtmodel)
        if len(self.get_future_hours()) >= 24:
            set_scooped_permittance(
                self.get_future_hours(), self.options.cautionhour_type_enum
            )

    async def async_update(self):
        self.update()

    def get_average_kwh_price(self) -> float:
        try:
            return round(
                mean(
                    [
                        hp.permittance * hp.price
                        for hp in self.future_hours
                        if hp.permittance > 0
                    ]
                ),
                2,
            )
        except Exception:
            return 0.0

    def get_future_hours(self) -> list[HourPrice]:
        ret = [hp for hp in self.model.hours_prices if not hp.passed]
        if self.max_min.active:
            for r in ret:
                if r in self.max_min.non_hours:
                    r.permittance = 0.0
                elif r.dt in self.max_min.dynamic_caution_hours.keys():
                    r.permittance = self.max_min.dynamic_caution_hours[r.dt]
        return ret

    async def async_update_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ):
        self.model.prices_today = prices  # clean first
        self.model.prices_tomorrow = prices_tomorrow  # clean first
        if self._do_recalculate_prices(prices, prices_tomorrow):
            await self.async_create_prices(prices, prices_tomorrow)

    async def async_update_adjusted_average(self, adjusted_average: float) -> None:
        self.model.adjusted_average = adjusted_average
        self.update()
        if len(self.model.hours_prices) > 0:
            self._set_permittance()

    def _do_recalculate_prices(self, prices, prices_tomorrow) -> bool:
        if [
            hp.price
            for hp in self.model.hours_prices
            if hp.dt.date() == self.dtmodel.hdate
        ] == prices and len(prices_tomorrow) < 1:
            return False
        return True

    async def async_create_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ) -> None:
        # todo: fix to allow 23, 24,25, 92, 96, 100 for dst-dates.
        self.model.hours_prices = []
        match len(prices):
            case 23 | 24 | 25:
                await self.async_create_hour_prices(
                    prices, prices_tomorrow, is_quarterly=False
                )
            case 92 | 96 | 100:
                await self.async_create_hour_prices(
                    prices, prices_tomorrow, is_quarterly=True
                )
            case _:
                raise ValueError(
                    f"Length of pricelist must be either 23,24,25,92,96 or 100. Your length is {len(prices)}"
                )

    async def async_create_hour_prices(
        self,
        prices: list[float],
        prices_tomorrow: list[float] = [],
        is_quarterly: bool = False,
    ) -> None:
        # todo: handle here first if prices or prices_tomorrow are 92 or 100 in length (dst shift)
        self.use_quarters = is_quarterly
        self.model.set_hourprice_list(
            prices,
            self.options,
            is_quarterly,
            self.dtmodel.hdate,
            self.dtmodel.is_passed,
        )
        self.model.set_hourprice_list(
            prices_tomorrow,
            self.options,
            is_quarterly,
            self.dtmodel.hdate_tomorrow,
            self.dtmodel.is_passed,
        )
        self._set_permittance()

    def _set_permittance(self) -> None:
        prices = normalize_prices([hp.price for hp in self.model.hours_prices])
        set_initial_permittance(
            self.model.hours_prices,
            mean(prices),
            stdev(prices),
            self.model.adjusted_average,
        )
        set_scooped_permittance(
            self.model.hours_prices,
            self.options.cautionhour_type_enum,
        )
        self._offset_dict = set_offset_dict(
            prices, self.model.hours_prices[0].dt.date()
        )
        self._block_nocturnal(self.model.hours_prices, self.options.blocknocturnal)

    @staticmethod
    def _block_nocturnal(
        hour_prices: list[HourPrice], block_nocturnal: bool = False
    ) -> None:
        blockhours = [23, 0, 1, 2, 3, 4, 5, 6]
        if block_nocturnal:
            for hp in hour_prices:
                if hp.hour in blockhours:
                    hp.permittance = 0.0



from ...models.locale.enums.calendar_periods import CalendarPeriods
from typing import Dict, List
from dataclasses import dataclass
from datetime import datetime

TIMEPATTERN_FUNC = {
    CalendarPeriods.Hour: lambda a, dtp: dtp.hour in a,
    CalendarPeriods.Weekday: lambda a, dtp: dtp.weekday() in a,
    CalendarPeriods.Month: lambda a, dtp: dtp.month in a,
}


@dataclass
class TimePattern:
    pattern: List[Dict[CalendarPeriods, List[int]]]

    def valid(self, dt=datetime.now()) -> bool:
        total = []
        for p in self.pattern:
            total.append(all(TIMEPATTERN_FUNC[k](v, dt) for k, v in p.items()))
        return any(total)

    async def async_valid(self, dt=datetime.now()) -> bool:
        total = []
        for p in self.pattern:
            total.append(all(TIMEPATTERN_FUNC[k](v, dt) for k, v in p.items()))
        return any(total)



import logging
from dataclasses import dataclass, field
from typing import List, Dict
from .hourobjects.hourobject import HourObject
from .hourtypelist import HourTypeList

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=False)
class HoursModel:
    non_hours: List[int] = field(default_factory=lambda : [])
    caution_hours: List[int] = field(default_factory=lambda : [])
    dynamic_caution_hours: Dict[int, float] = field(default_factory=lambda : {})
    hours_today: HourObject = field(default_factory=lambda : HourObject([],[],dict()))
    hours_tomorrow: HourObject = field(default_factory=lambda : HourObject([],[],dict()))
    offset_dict: Dict[Dict[str,float], Dict[str, float]] = field(default_factory=lambda: {})

    async def async_update_non_hours(self,hour:int) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.nh if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.nh if h < hour)
        self.non_hours = ret
    
    def _update_non_hours(self,hour:int) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.nh if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.nh if h < hour)
        self.non_hours = ret

    async def async_update_caution_hours(self, hour:int) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.ch if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.ch if h < hour)
        self.caution_hours = ret

    def _update_caution_hours(self, hour:int) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.ch if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.ch if h < hour)
        self.caution_hours = ret

    async def async_update_dynamic_caution_hours(self, hour:int) -> None:
        ret = {}
        ret.update({k: v for k, v in self.hours_today.dyn_ch.items() if k >= hour and k not in self.hours_today.nh})
        ret.update({k: v for k, v in self.hours_tomorrow.dyn_ch.items() if k < hour and k not in self.hours_tomorrow.nh})
        self.dynamic_caution_hours = ret

    def _update_dynamic_caution_hours(self, hour:int) -> None:
        ret = {}
        ret.update({k: v for k, v in self.hours_today.dyn_ch.items() if k >= hour and k not in self.hours_today.nh})
        ret.update({k: v for k, v in self.hours_tomorrow.dyn_ch.items() if k < hour and k not in self.hours_tomorrow.nh})
        self.dynamic_caution_hours = ret

    async def async_update_offset_dict(self) -> None:
        ret = {}
        ret['today'] = self.hours_today.offset_dict
        ret['tomorrow'] = self.hours_tomorrow.offset_dict
        self.offset_dict = ret

    async def async_touch_midnight(self) -> bool:
        self.hours_today = self.hours_tomorrow
        self.hours_tomorrow = HourObject([], [], {})
        self.offset_dict["today"] = self.offset_dict.get("tomorrow", {})
        self.offset_dict["tomorrow"] = {}
        return True

    async def async_update_hour_lists(self, hour:int) -> None:
        await self.async_update_non_hours(hour)
        await self.async_update_caution_hours(hour)
        await self.async_update_dynamic_caution_hours(hour)
        await self.async_update_offset_dict()

    def update_hour_list(self, hour:int, listtype:HourTypeList = None) -> None:
        match listtype:
            case HourTypeList.NonHour:
                self._update_non_hours(hour)
            case HourTypeList.CautionHour:
                self._update_caution_hours(hour)
            case HourTypeList.DynCautionHour:
                self._update_dynamic_caution_hours(hour)   
            case _:
                pass


import logging

from ...util import nametoid
from ...models.const import DOMAIN
from .hubmember import HubMember
from ...hub.killswitch import KillSwitch
from .const import TOTALPOWER, HOUSEPOWER

_LOGGER = logging.getLogger(__name__)


class Power:
    def __init__(self, configsensor: str, powersensor_includes_car: bool = False):
        self._config_sensor = configsensor
        self._total = HubMember(data_type=int, initval=0, name=TOTALPOWER)
        self._house = HubMember(data_type=int, initval=0, name=HOUSEPOWER)
        self._powersensor_includes_car = powersensor_includes_car
        self.killswitch = KillSwitch(
            sensor=self._config_sensor, update_interval=120, grace_interval=300
        )
        self._setup()

    @property
    def is_initialized(self) -> bool:
        return self._total.is_initialized and self._house.is_initialized

    @property
    def config_sensor(self) -> str:
        return self._config_sensor

    @property
    def total(self) -> HubMember:
        self.killswitch.check
        return self._total

    @total.setter
    def total(self, val):
        self._total = val

    @property
    def house(self) -> HubMember:
        self.killswitch.check
        return self._house

    @house.setter
    def house(self, val):
        self._house = val

    @property
    def car_power(self) -> int:
        try:
            ret = int(self._total.value - self._house.value)
            return max(0, ret)
        except:
            return 0

    def _setup(self):
        if self._powersensor_includes_car is True:
            self.total.entity = self.config_sensor
            self.house.entity = nametoid(f"sensor.{DOMAIN}_{HOUSEPOWER}")
        else:
            self.house.entity = self._config_sensor

    async def async_update(self, carpowersensor_value=0, config_sensor_value=None):
        self.update(carpowersensor_value, config_sensor_value)
        # if self._powersensor_includes_car is True:
        #     if config_sensor_value is not None:
        #         self.total.value = config_sensor_value
        #     new_val = (float(self.total.value) - float(carpowersensor_value))
        #     if new_val != self.house.value:
        #         await self.killswitch.async_update()
        #     self.house.value = new_val
        # else:
        #     if config_sensor_value is not None:
        #         self.house.value = config_sensor_value
        #     new_val = (float(self.house.value) + float(carpowersensor_value))
        #     if new_val != self.total.value:
        #         await self.killswitch.async_update()
        #     self.total.value = new_val

    def update(self, carpowersensor_value=0, config_sensor_value=None):
        if self._powersensor_includes_car is True:
            if config_sensor_value is not None:
                self.total.value = config_sensor_value
            new_val = float(self.total.value) - float(carpowersensor_value)
            if new_val != self.house.value:
                self.killswitch.update()
            self.house.value = new_val
        else:
            if config_sensor_value is not None:
                self.house.value = config_sensor_value
            new_val = float(self.house.value) + float(carpowersensor_value)
            if new_val != self.total.value:
                self.killswitch.update()
            self.total.value = new_val
